#pragma once

//Streams
#include <iostream>
#include <fstream>

//For random numbers generator
#include <cstdlib>
#include <ctime>
#include <random>

#include <algorithm>
#include <iterator>
#include <ranges>
#include <numeric>

#include "Exception.h"

//Sequence containers
#include <vector>
#include <list>
#include <deque>


//=================================================================


//Creates text file with 100 random integers in range [-50, 50]
std::fstream createData()
{
	std::fstream file("100_Random_Nums.txt", std::ios::out | std::ios::trunc);
	

	std::srand(unsigned(std::time(nullptr)));

	if (!file.is_open())
	{
		throw MyException("Failure. File was not opened...\n");
	}
	else
	{
		for (int i : std::ranges::views::iota(0, 100))
		{
			file << std::rand() % 101 - 50 << " ";
		}

		file.close();

		return file;
	}
}


//Creates text file with 100 random integers in range [-50, 50] with recording in container
template<typename _ContType>
std::fstream createData(_ContType& cont)
{
	std::fstream file("100_Random_Nums.txt", std::ios::out | std::ios::trunc);
	std::ostream_iterator<_ContType::value_type> file_iter(file, " ");

	if (!file.is_open())
	{
		throw MyException("Failure. File was not opened...\n");
	}
	else
	{
		for (int i : std::ranges::views::iota(-50, 51))
		{
			*std::back_inserter(cont) = i;
		}

		//Shuffle container elements
		std::mt19937 gen{ std::random_device() };
		std::ranges::shuffle(cont, gen);


		std::ranges::copy(cont, file_iter);

		file.close();

		return file;
	}
}


//=================================================================


//Reads data from file into sequence container(usage of file streams)
template<typename _ContType>
_ContType loadData(const char* filename = "100_Random_Nums.txt")
{
	_ContType cont;
	std::ifstream file(filename, std::ios::in);

	if (!file.is_open())
	{
		throw MyException("Failure.File was not opened...\n");
	}
	else
	{
		char* cur_num = new char[4];
		file.seekg(0);
		while (!file.eof())
		{
			cur_num[0] = '\n';
			file.getline(cur_num, 4, ' ');
			*std::back_inserter(cont) = atoi(cur_num);
		}

		cont.pop_back(); //      

		file.close();

		return cont;
	}
}


//Reads data from file into given sequence container(usage of stream iterators)
template<typename _ContType>
_ContType& loadData(_ContType& cont, const char* filename = "100_Random_Nums.txt")
{
	std::ifstream file(filename, std::ios::in);

	if (!file.is_open())
	{
		throw MyException("Failure.File was not opened...\n");
	}
	else
	{
		cont.clear();
		file.seekg(0); //rewind
		std::ranges::copy(file, std::back_inserter(cont));

		file.close();

		return cont;
	}

}


//=================================================================


//Modifying of SEQUENCE container(multiplying of every container element on minimal element in container
//and adding of average of all container elements to the end of container)
template<typename _ContType>
void modify(const _ContType& cont)
{
	if (cont.size() == 0)
		throw MyException("\nProgram cannot modify elements of an empty container!\n");
	
	typename _ContType::value_type& min_el = std::ranges::min_element(cont, std::less<>); //minimal element in container
	std::ranges::for_each(cont, [&min_el](_ContType::value_type& cur_container_el) {cur_container_el *= min_el; });
	
	typename _ContType::value_type sum = std::accumulate(std::begin(cont), std::end(cont), 0, std::plus<_ContType::value_type>);
	*std::back_inserter(cont) = std::fabs(sum * 1.0 / cont.size());
}


//Modifying of range contained in SEQUENCE container(multiplying of every range elements on minimal element in this range)
template<typename _IterType>
void modify(const _IterType beg, const _IterType end)
{
	if (std::distance(beg, end) <= 0)
		throw MyException("Iterators of diapazon are not in valid sequence!");

	typedef typename std::iterator_traits<_IterType>::value_type value_type; //definition of type na kotoriy ukazivaet iterator with type "_IterType"

	auto min_el = std::ranges::min_element(beg, end, std::less<value_type>); //minimal element in container
	std::ranges::for_each(beg, end, [&min_el](value_type& cur_el) {cur_el *= min_el; });
}


std::fstream& modify(const char* filename = "100_Random_Nums.txt")
{
	std::fstream file(filename, std::ios::in);

	if (!file.is_open())
		throw MyException("Failed to load file for modification");

	auto min_el = std::min_element(std::istream_iterator<int>(file), std::istream_iterator<int>(), std::less<int>());
	std::transform(std::istream_iterator<int>(file), std::istream_iterator<int>(),
						   std::ostream_iterator<int>(file, " "),
						   [min_el](std::iterator_traits<std::istream_iterator<int>>::value_type cur_num) {return cur_num * (*min_el); });

	file.close();

	return file;
}


//=================================================================


//
template<typename _ContType>
std::fstream outputResult(const _ContType& cont)
{
	std::ofstream outp_file("Result_File.txt", std::ios::out | std::ios::trunc);

	if (!outp_file.is_open())
		throw MyException("Error! Failure in opening output file.");

	outp_file.seekp(0);

	for (auto cur_el : cont)
	{
		outp_file << cur_el << " ";
		if ((std::distance(*cur_el, cont.begin()) > 0) && (std::distance(*cur_el, cont.begin()) % 10 == 0))
			outp_file << "\n";
	}

	outp_file.close();

	return outp_file;
}


template<typename _ContType>
std::fstream outputResult(const char* outp_file_name, const _ContType& cont)
{
	std::ofstream outp_file(outp_file_name, std::ios::out | std::ios::trunc);
	std::ostream_iterator file_iter(outp_file, " ");


	if (!outp_file.is_open())
		throw MyException("Error! Failure in opening output file.");

	std::ranges::copy(cont, file_iter);

	outp_file.close();

	return outp_file;
}


/*
template<typename _DataType>
std::deque<_DataType> loadData(std::fstream& file)
{
	std::deque<_DataType> tmp_deque;
	_DataType cur_elem;
	
	file.seekp(0, std::ios::beg);
	while (file >> cur_elem)
	{
		if (cur_elem % 2 == 0)
		{
			tmp_deque.push_back(cur_elem);
		}
		else
		{
			tmp_deque.push_front(cur_elem);
		}
	}

	return tmp_deque;
}
*/

/*
//Reads data from file into deque with given condition
template<typename _DataType, typename _Pred>
std::deque<_DataType> loadData(std::fstream& file, _Pred cond)
{
	std::deque<_DataType> tmp_deque;
	_DataType cur_elem;

	file.seekp(0, std::ios::beg);
	while (file >> cur_elem)
	{
		if (cond(cur_elem))
		{
			tmp_deque.push_back(cur_elem);
		}
		else
		{
			tmp_deque.push_front(cur_elem);
		}
	}

	return tmp_deque;
}
*/


